' table.Event.vb
'
' Copyright (c) 2008-2011, SystemBase Co.,Ltd.
' All rights reserved.
'
' Redistribution and use in source and binary forms, with or without
' modification, are permitted provided that the following conditions are met:
'
'    1. Redistributions of source code must retain the above copyright
'       notice, this list of conditions and the following disclaimer.
'    2. Redistributions in binary form must reproduce the above copyright
'       notice, this list of conditions and the following disclaimer in the
'       documentation and/or other materials provided with the distribution.
'
' THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
' IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
' ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
' LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
' CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
' SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
' INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
' CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
' ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
' POSSIBILITY OF SUCH DAMAGE.

Imports System.Drawing

Partial Class UTable

    Private Sub UTable_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Click
        Me.Focus()
    End Sub

    Private Sub UTable_Enter(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Enter
        Me.Focus()
    End Sub

    Protected Overrides Function IsInputChar(ByVal charCode As Char) As Boolean
        Return True
    End Function

    Protected Overrides Function IsInputKey(ByVal keyData As System.Windows.Forms.Keys) As Boolean
        If Me.Editor Is Nothing Then
            Return Me.KeyboardOperation.IsInputKey(keyData, Me, MyBase.IsInputKey(keyData))
        End If
        Return MyBase.IsInputKey(keyData)
    End Function

    Protected Overrides Function ProcessDialogKey(ByVal keyData As System.Windows.Forms.Keys) As Boolean
        If Me.Editor Is Nothing Then
            If Me.KeyboardOperation.ProcessDialogKey(keyData, Me) Then
                Return True
            End If
        End If
        Return MyBase.ProcessDialogKey(keyData)
    End Function

    Public MouseDownField As CField = Nothing
    Private Sub UTable_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDown
        If Me.StrictMouseLocation AndAlso _
           Not Me.PointToClient(Cursor.Position).Equals(e.Location) Then
            Exit Sub
        End If
        If Me.Editor IsNot Nothing Then
            If Not Me.LeaveEdit Then
                Exit Sub
            End If
        End If
        If Me.cols_MouseDown(e) Then
            Exit Sub
        End If
        If Me.focus_MouseDown(e) Then
            Exit Sub
        End If
        If Me.Editor Is Nothing Then
            Dim field As CRenderCache.CField = Me.RenderCache.FindField(e.Location)
            If field IsNot Nothing Then
                Me.MouseDownField = field.Field
                Dim l As New Point(e.X - field.Rect.Left, e.Y - field.Rect.Top)
                Me.recordDrag_MouseDown(field.Field, l, e)
                field.Field.raiseMouseDown(l, e)
                RaiseEvent FieldMouseDown(field.Field, l, e)
                Exit Sub
            End If
        End If
    End Sub

    Private Sub UTable_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseUp
        If Me.cols_MouseUp(e) Then
            Exit Sub
        End If
        If Me.range_MouseUp(e) Then
            Exit Sub
        End If
        If Me.Editor Is Nothing Then
            Me.recordDrag_MouseUp(e)
            If Me.MouseDownField IsNot Nothing Then
                Dim field As CRenderCache.CField = Me.RenderCache.FindField(e.Location)
                Dim l As Point
                If field IsNot Nothing AndAlso Me.MouseDownField Is field.Field Then
                    l = New Point(e.X - field.Rect.Left, e.Y - field.Rect.Top)
                Else
                    l = New Point(-1, -1)
                End If
                Dim _f = Me.MouseDownField
                Me.MouseDownField = Nothing
                _f.raiseMouseUp(l, e)
                RaiseEvent FieldMouseUp(_f, l, e)
                Exit Sub
            End If
        End If
    End Sub

    Private Sub UTable_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseMove
        If Me.StrictMouseLocation AndAlso _
           Not Me.PointToClient(Cursor.Position).Equals(e.Location) Then
            Exit Sub
        End If
        If Me.cols_MouseMove(e) Then
            Exit Sub
        End If
        If Me.drag_MouseMove(e) Then
            Exit Sub
        End If
        If Me.range_MouseMove(e) Then
            Exit Sub
        End If
        If Me.Editor Is Nothing Then
            Me.recordDrag_MouseMove(e)
            Dim field As CRenderCache.CField = Me.RenderCache.FindField(e.Location)
            If field IsNot Nothing Then
                Dim l As New Point(e.X - field.Rect.Left, e.Y - field.Rect.Top)
                field.Field.raiseMouseMove(l, e)
                RaiseEvent FieldMouseMove(field.Field, l, e)
                Exit Sub
            End If
        End If
    End Sub

    Private Sub UTable_MouseClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseClick
        If Me.StrictMouseLocation AndAlso _
           Not Me.PointToClient(Cursor.Position).Equals(e.Location) Then
            Exit Sub
        End If
        If Me.Editor Is Nothing Then
            Dim field As CRenderCache.CField = Me.RenderCache.FindField(e.Location)
            If field IsNot Nothing Then
                Dim l As New Point(e.X - field.Rect.Left, e.Y - field.Rect.Top)
                field.Field.raiseClick(l, e)
                RaiseEvent FieldClick(field.Field, l, e)
                Exit Sub
            End If
        End If
    End Sub

    Public MouseDoubleClicking As Boolean = False
    Private Sub UTable_MouseDoubleClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseDoubleClick
        Me.MouseDoubleClicking = True
        Try
            If Me.StrictMouseLocation AndAlso _
               Not Me.PointToClient(Cursor.Position).Equals(e.Location) Then
                Exit Sub
            End If
            If Me.cols_MouseDoubleClick(e) Then
                Exit Sub
            End If
            If Me.edit_MouseDoubleClick(e) Then
                Exit Sub
            End If
            If Me.Editor Is Nothing Then
                Dim field As CRenderCache.CField = Me.RenderCache.FindField(e.Location)
                If field IsNot Nothing Then
                    Dim l As New Point(e.X - field.Rect.Left, e.Y - field.Rect.Top)
                    field.Field.raiseDoubleClick(l, e)
                    RaiseEvent FieldDoubleClick(field.Field, l, e)
                    Exit Sub
                End If
            End If
        Finally
            Me.MouseDoubleClicking = False
        End Try
    End Sub

    Private Sub UTable_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyDown
        If Me.edit_KeyDown(e) Then
            Exit Sub
        End If
        If Me.Editor Is Nothing Then
            If Me.FocusField IsNot Nothing Then
                Me.FocusField.raiseKeyDown(e)
                RaiseEvent FieldKeyDown(Me.FocusField, e)
                Exit Sub
            End If
        End If
    End Sub

    Private Sub UTable_KeyUp(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles Me.KeyUp
        If Me.Editor Is Nothing Then
            If Me.FocusField IsNot Nothing Then
                Me.FocusField.raiseKeyUp(e)
                RaiseEvent FieldKeyUp(Me.FocusField, e)
                Exit Sub
            End If
        End If
    End Sub

    Private Sub UTable_KeyPress(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyPressEventArgs) Handles Me.KeyPress
        If Me.imeSetUpFlag Then
            Exit Sub
        End If
        If Me.edit_KeyPress(e) Then
            Exit Sub
        End If
        If Me.Editor Is Nothing Then
            If Me.FocusField IsNot Nothing Then
                Me.FocusField.raiseKeyPress(e)
                RaiseEvent FieldKeyPress(Me.FocusField, e)
            End If
        End If
    End Sub

    Private Sub UTable_GotFocus(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.GotFocus
        Using Me.RenderBlock
            If Me.Editor IsNot Nothing Then
                Me.Editor.Control.Focus()
            End If
            If Me.FocusField IsNot Nothing Then
                Me.imeSetUp(Me.FocusField)
            End If
            If Me.Setting.EnterContent IsNot Nothing AndAlso Me.FocusField Is Nothing Then
                Me.FocusField = Me.Setting.EnterContent.DefaultField
            End If
        End Using
    End Sub

    Private Sub UTable_LostFocus(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.LostFocus
        Me.range_LostFocus(e)
        Me.Render()
    End Sub

    Private Sub UTable_Resize(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Resize
        Dim context As BufferedGraphicsContext = BufferedGraphicsManager.Current
        Try
            Using Me.RenderBlock
                If Me.Editor IsNot Nothing Then
                    If Not Me.LeaveEdit Then
                        Me.FinishEdit()
                    End If
                End If
                Me.BackBuffer = context.Allocate(Me.CreateGraphics, Me.ClientRectangle)
                Me.LayoutCacheInvalid = True
            End Using
        Catch ex As Exception
        End Try
    End Sub

    Private Sub UTable_DraggingScroll(ByVal table As UTable, ByVal dx As Integer, ByVal dy As Integer) Handles Me.DraggingScroll
        If Me.range_DraggingOverflow(dx, dy) Then
            Exit Sub
        End If
        Me.recordDrag_DraggingOverflow(dx, dy)
    End Sub

    Private Class CDraggingScrollMove
        Public dx As Integer = 0
        Public dy As Integer = 0
        Private Const THRESHOLD As Integer = 80
        Public Sub Update(ByVal table As UTable, ByVal location As Point)
            If location.X < -THRESHOLD Then
                dx = -10
            ElseIf location.X < 0 Then
                dx = -1
            ElseIf location.X > table.Width + THRESHOLD Then
                dx = 10
            ElseIf location.X > table.Width Then
                dx = 1
            Else
                dx = 0
            End If
            If location.Y < table.HeaderContent.LayoutCache.RowsSize - THRESHOLD Then
                dy = -10
            ElseIf location.Y < table.HeaderContent.LayoutCache.RowsSize Then
                dy = -1
            ElseIf location.Y > table.Height - table.FooterContent.LayoutCache.RowsSize + THRESHOLD Then
                dy = 10
            ElseIf location.Y > table.Height - table.FooterContent.LayoutCache.RowsSize Then
                dy = 1
            Else
                dy = 0
            End If
        End Sub
    End Class

    Private draggingScrollMove As New CDraggingScrollMove
    Private WithEvents draggingScrollTimer As New Timer

    Private Sub draggingScrollTimer_Tick(ByVal sender As Object, ByVal e As System.EventArgs) Handles draggingScrollTimer.Tick
        With Me.draggingScrollMove
            If .dx <> 0 Or .dy <> 0 Then
                RaiseEvent DraggingScroll(Me, .dx, .dy)
            End If
        End With
    End Sub

    Private Function drag_MouseMove(ByVal e As MouseEventArgs) As Boolean
        Me.draggingScrollMove.Update(Me, e.Location)
        Return False
    End Function

    Private Sub UTable_Validating(ByVal sender As Object, ByVal e As System.ComponentModel.CancelEventArgs) Handles Me.Validating
        If Me.Editor IsNot Nothing Then
            e.Cancel = True
        End If
    End Sub

End Class
