' table/Grid.vb
'
' Copyright (c) 2008-2011, SystemBase Co.,Ltd.
' All rights reserved.
'
' Redistribution and use in source and binary forms, with or without
' modification, are permitted provided that the following conditions are met:
'
'    1. Redistributions of source code must retain the above copyright
'       notice, this list of conditions and the following disclaimer.
'    2. Redistributions in binary form must reproduce the above copyright
'       notice, this list of conditions and the following disclaimer in the
'       documentation and/or other materials provided with the distribution.
'
' THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
' IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
' ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
' LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
' CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
' SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
' INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
' CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
' ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
' POSSIBILITY OF SUCH DAMAGE.

Imports System.Drawing

Partial Class UTable

    Public Class CGrid

        Public Class CPoint
            Public Row As Integer
            Public Col As Integer
            Public Sub New()
                Me.New(0, 0)
            End Sub
            Public Sub New(ByVal row As Integer, ByVal col As Integer)
                Me.Row = row
                Me.Col = col
            End Sub
        End Class

        Public Class CSize
            Public Row As Integer
            Public Col As Integer
            Public Sub New()
                Me.New(0, 0)
            End Sub
            Public Sub New(ByVal row As Integer, ByVal col As Integer)
                Me.Row = row
                Me.Col = col
            End Sub
        End Class

        Public Class CRegion
            Public Row As Integer
            Public Col As Integer
            Public Rows As Integer
            Public Cols As Integer

            Public Sub New()
                Me.New(0, 0)
            End Sub

            Public Sub New(ByVal row As Integer, ByVal col As Integer)
                Me.New(row, col, 1, 1)
            End Sub

            Public Function Clone() As CRegion
                Dim ret As New CRegion
                ret.Row = Me.Row
                ret.Col = Me.Col
                ret.Rows = Me.Rows
                ret.Cols = Me.Cols
                Return ret
            End Function

            Public Sub New(ByVal row As Integer, _
                           ByVal col As Integer, _
                           ByVal rows As Integer, _
                           ByVal cols As Integer)
                Me.Row = row
                Me.Col = col
                Me.Rows = rows
                Me.Cols = cols
            End Sub

            Public Function Contains(ByVal point As CPoint) As Boolean
                Return point.Row >= Me.Row And point.Row <= Me.BottomRow And _
                       point.Col >= Me.Col And point.Col <= Me.RightCol
            End Function

            Public Function Point() As CPoint
                Return New CPoint(Me.Row, Me.Col)
            End Function

            Public Function BottomRow() As Integer
                Return (Me.Row + Me.Rows - 1)
            End Function

            Public Function RightCol() As Integer
                Return (Me.Col + Me.Cols - 1)
            End Function

            Public Function Union(ByVal r As CRegion) As CRegion
                Dim row As Integer = 0
                Dim col As Integer = 0
                Dim rows As Integer = 0
                Dim cols As Integer = 0
                If Me.Row < r.Row Then row = Me.Row Else row = r.Row
                If Me.Col < r.Col Then col = Me.Col Else col = r.Col
                If Me.Row + Me.Rows > r.Row + r.Rows Then
                    rows = Me.Row + Me.Rows - row
                Else
                    rows = r.Row + r.Rows - row
                End If
                If Me.Col + Me.Cols > r.Col + r.Cols Then
                    cols = Me.Col + Me.Cols - col
                Else
                    cols = r.Col + r.Cols - col
                End If
                Return New CRegion(row, col, rows, cols)
            End Function

        End Class

        Public Table As UTable = Nothing
        Public Record As UTable.CRecord = Nothing
        Public Content As UTable.CContent = Nothing
        Public SizeMax As Integer = 0
        Public SizeMin As Integer = 0        
        Public UserResizable As EAllow = EAllow.DEFAULT
        Public UserAdjustable As EAllow = EAllow.DEFAULT

        Private _Visible As Boolean = True
        Private _Size As Integer = 0

        Public Sub New()
        End Sub

        Public Sub New(ByVal table As UTable)
            Me.Table = table
            Me.Record = Nothing
            Me.Content = Nothing
        End Sub

        Public Sub New(ByVal record As CRecord)
            Me.Table = Nothing
            Me.Record = record
            Me.Content = record.TopLevelContent
        End Sub

        Public Sub New(ByVal i As CGrid, ByVal record As CRecord)
            Me.Table = Nothing
            Me.Record = record
            Me.Content = record.TopLevelContent
            Me.SizeMin = i.SizeMin
            Me.SizeMax = i.SizeMax
            Me.UserResizable = i.UserResizable
            Me.UserAdjustable = i.UserAdjustable
            Me.Visible = i.Visible
            Me.Size = i.Size
        End Sub

        Public Shared Function GetRectangle(ByVal cols As CGrids, _
                                            ByVal rows As CGrids, _
                                            ByVal region As CRegion, _
                                            ByVal top_row As Integer, _
                                            ByVal top_col As Integer) As Rectangle
            Dim row_pos As Integer
            Dim row_size As Integer
            Dim col_pos As Integer
            Dim col_size As Integer

            row_size = rows.GetSize(region.Row, region.Rows)

            If row_size = 0 Then
                Return Nothing
            End If

            col_size = cols.GetSize(region.Col, region.Cols)

            If col_size = 0 Then
                Return Nothing
            End If

            row_pos = top_row + rows.GetLocation(region.Row)
            col_pos = top_col + cols.GetLocation(region.Col)

            Return New Rectangle(col_pos, row_pos, col_size, row_size)

        End Function

        Public Property Size() As Integer
            Get
                If Me._Size = 0 Then
                    If Me.Table IsNot Nothing Then
                        Return Me.Table.DefaultGridSize.Col
                    ElseIf Me.Content IsNot Nothing Then
                        Return Me.Content.Table.DefaultGridSize.Row
                    Else
                        Return 0
                    End If
                Else
                    Return Me._Size
                End If
            End Get
            Set(ByVal value As Integer)
                Dim v As Integer = Me.Regularize(value)
                If v > 0 Then
                    Me._Size = v
                    Me.update()
                End If
            End Set
        End Property

        Public Property Visible() As Boolean
            Get
                Return Me._Visible
            End Get
            Set(ByVal value As Boolean)
                Me._Visible = value
                Me.update()
            End Set
        End Property

        Public Function Regularize(ByVal size As Integer) As Integer
            If Me.SizeMin > 0 AndAlso size < Me.SizeMin Then
                Return Me.SizeMin
            ElseIf Me.SizeMax > 0 AndAlso size > Me.SizeMax Then
                Return Me.SizeMax
            Else
                Return size
            End If
        End Function

        Private Sub update()
            If Me.Table IsNot Nothing Then
                Using Me.Table.RenderBlock
                    Me.Table.LayoutCacheInvalid = True
                End Using
            ElseIf Me.Content IsNot Nothing Then
                Using Me.Content.Table.RenderBlock
                    Me.Content.TopLevelContent.LayoutCacheInvalid = True
                End Using
            End If
        End Sub

        Public Function Index() As Integer
            If Me.Table IsNot Nothing Then
                Return Me.Table.Cols.IndexOf(Me)
            Else
                Return Me.Record.Rows.IndexOf(Me)
            End If
        End Function

    End Class

    Public Class CGrids
        Inherits List(Of CGrid)

        Public Sub New()
        End Sub

        Public Sub New(ByVal grids As CGrids, ByVal record As CRecord)
            For Each grid As CGrid In grids
                Me.Add(New CGrid(grid, record))
            Next
        End Sub

        Public Function GetLocation(ByVal pos As Integer) As Integer
            Dim ret As Integer = 0
            For i As Integer = 0 To pos - 1
                If Me(i).Visible Then
                    ret += Me(i).Size
                End If
            Next
            Return ret
        End Function

        Public Function GetSize(ByVal pos As Integer, ByVal size As Integer) As Integer
            Dim ret As Integer = 0
            For i As Integer = 0 To size - 1
                If Me(pos + i).Visible Then
                    ret += Me(pos + i).Size
                End If
            Next
            Return ret
        End Function

        Public Sub SetSize(ByVal ParamArray v() As Integer)
            For i As Integer = 0 To v.Length - 1
                If i < Me.Count Then
                    Me(i).Size = v(i)
                End If
            Next
        End Sub

        Public Function GetSizeByEntireSize(ByVal region_start As Integer, _
                                            ByVal region_size As Integer, _
                                            ByVal position As Integer, _
                                            ByVal entireSize As Integer) As Integer
            Return entireSize - (Me.GetSize(region_start, region_size) - Me.GetSize(position, 1)) + 2
        End Function

    End Class

    Public Function PosToCol(ByVal x As Integer) As Integer
        If x < Me.LayoutCache.FixedColsSize Then
            Dim _x As Integer = x
            Dim w As Integer = 0
            For i As Integer = 0 To Me.FixedCols - 1
                If Me.Cols(i).Visible Then
                    w += Me.Cols(i).Size
                End If
                If _x < w Then
                    Return i
                End If
            Next
        Else
            Dim _x As Integer = x + Me.HScrollBar.Value - Me.LayoutCache.FixedColsSize
            Dim w As Integer = 0
            For i As Integer = Me.FixedCols To Me.Cols.Count - 1
                If Me.Cols(i).Visible Then
                    w += Me.Cols(i).Size
                End If
                If _x < w Then
                    Return i
                End If
            Next
        End If
        Return -1
    End Function

End Class